/*
 * tcp.c -- telnet protocol communication module for powwow
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/telnet.h>
#ifndef TELOPT_NAWS
#  define TELOPT_NAWS 31
#endif
#include <arpa/inet.h>
#ifndef NEXT
#  include <unistd.h>
#endif

#ifdef TERM
#  include "client.h"
#endif

#include "powwow.h"
#include "tcp.h"
#include "tty.h"

#ifndef OSF1
unsigned long inet_addr();
#endif

#ifdef TELOPTS
# define TELOPTSTR(n) ((n) > NTELOPTS ? "unknown" : telopts[n])
#endif

static void dosubopt();
static void sendopt();

static int sending_size = 0;	/* true if we're communicating window size */

int tcp_fd = -1;                   /* current socket file descriptor
				      -1 means no socket */
int tcp_main_fd = -1;		   /* socket file descriptor of main connect.
				      -1 means no socket */
int tcp_count = 0;		   /* number of open connections */

connsess conn_list[MAXFDSCAN];     /* connections indexed by socket descrs */
#define CONN_LIST(n) conn_list[n]

fd_set fdset;			/* set of descriptors to select() on */


/*
 * connect to remote host
 * Warning: some voodoo code here
 */
int tcp_connect(addr, port)
char *addr; int port;
{
    struct sockaddr_in address;
    struct hostent *host_info;
    int err, newtcp_fd;
    
    bzero((char *)&address, sizeof(address));
    /*
     * inet_addr has a strange design: It is documented to return -1 for
     * malformed requests, but it is declared to return unsigned long!
     * Anyway, this works.
     */    
    
#ifndef TERM  
    address.sin_addr.s_addr = inet_addr(addr);
    if (address.sin_addr.s_addr != (unsigned int)-1)
	address.sin_family = AF_INET;
    else
    {
	host_info = gethostbyname(addr);
	if (host_info == 0) {
	    PRINT_INT("#powwow: gethostbyname: unknown host\n");
	    return -1;
	}
	bcopy(host_info->h_addr, (char*)&address.sin_addr,
	      host_info->h_length);
	address.sin_family = host_info->h_addrtype;
    }
    address.sin_port = htons(port);
    
    newtcp_fd = socket(address.sin_family, SOCK_STREAM, 0);
    if (newtcp_fd < 0) {
	PRINT_INT("#powwow: error creating socket\n");
	return -errno;
    }
    
    PRINT_INT("#Trying %s... ", addr);
    tty_flush();
    
    err = connect(newtcp_fd, (struct sockaddr *)&address,	sizeof(address));
    
    if (err < 0 && errno == EINTR) /* CTRL-C pressed */
	return -errno;

#else /* term */
    
    if ((newtcp_fd = connect_server(0)) < 0)
    {
	tty_puts("\n#powwow: unable to connect to term server\n");
	return -1;
    }
    else
    {
	send_command(newtcp_fd, C_PORT, 0, "%s:%d", addr, port);
	tty_puts("Connected to term server...\n");
#ifdef TERM_COMPRESS
	send_command(newtcp_fd, C_COMPRESS, 1, "y");
#endif     
	send_command(newtcp_fd, C_DUMB, 1, 0);
    }
#endif /* term */
    
    if (err < 0 || newtcp_fd >= MAXFDSCAN) {
	close(newtcp_fd);
	return err < 0 ? -errno : -1;
    } else {
	tty_puts("connected!\n");
	return newtcp_fd;
    }
}

/*
 * read a maximum of size bytes from remote host
 * using the telnet protocol. return bytes read.
 */
int tcp_read(buffer, maxsize)
char *buffer; int maxsize;
{
    typedef enum {
	NORMAL = 0, GOTIAC, GOTWILL, GOTWONT, GOTDO, GOTDONT, GOTSB, GOTSBIAC
    } telnetstate;
    static telnetstate sock_state[MAXFDSCAN];
    telnetstate state = sock_state[tcp_fd];
    static unsigned char subopt[MAXSUBOPT];
    static int subchars;
    int i;
    unsigned char *p, *s;
    
    while ((i = read(tcp_fd, buffer, maxsize)) < 0 && errno == EINTR);
    
    if (i <= 0) {
	if (i == 0) {
	    sock_state[tcp_fd] = NORMAL;
	    tcp_close(NULL);
	    return 0;
	}
	syserr("read from socket");
    }
    /* scan through the buffer and interpret telnet protocol escapes */
    
    for (s = p = (unsigned char*)buffer; i--; s++) {
	switch(state) {
	 case NORMAL:
	 default:
	    if (*s != IAC)
		*p++ = *s;
	    else
		state = GOTIAC;
	    break;
	 case GOTIAC:
	    switch(*s) {
	     case WILL:
		state = GOTWILL; break;
	     case WONT:
		state = GOTWONT; break;
	     case DO:
		state = GOTDO; break;
	     case DONT:
		state = GOTDONT; break;
	     case SB:		/* BUG (multiple connections):	*/
		state = GOTSB;	/* there is only one subopt buffer */
		subchars = 0;
		break;
	     case IAC:
		*p++ = IAC;
		break;
	     case GA:
		/* I should handle GA as end-of-prompt marker one day */
	     default:
		/* ignore the rest of the telnet commands */
#ifdef TELOPTS
		tty_printf("[skipped IAC <%d>]\n", *s);
#endif
		state = NORMAL; break;
	    }
	    break;
	    
	 case GOTWILL:
#ifdef TELOPTS
	    tty_printf("[got WILL %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	     case TELOPT_ECHO:
		/* host echoes, turn off echo here
		 * but only for main connection, since we do not want
		 * subsidiary connection password entries to block anything
		 * in the main connection
		 */
		if (tcp_fd == tcp_main_fd)
		    linemode |= LM_NOECHO;
		sendopt(DO, *s);
		break;
	     case TELOPT_SGA:
		/* this can't hurt */
		linemode |= LM_CHAR;
		tty_special_keys();
		sendopt(DO, *s);
		break;
	     default:
		/* don't accept other options */
		sendopt(DONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;
	    
	 case GOTWONT:
#ifdef TELOPTS
	    tty_printf("[got WONT %s]\n", TELOPTSTR(*s));
#endif
	    if (*s == TELOPT_ECHO) {
		/* host no longer echoes, we do it instead */
		linemode &= ~LM_NOECHO;
	    }
	    /* accept any WONT */
	    sendopt(DONT, *s);
	    state = NORMAL;
	    break;
	    
	 case GOTDO:
#ifdef TELOPTS
	    tty_printf("[got DO %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	     case TELOPT_SGA:
		linemode |= LM_CHAR;
		tty_special_keys();
		/* FALLTHROUGH */
	     case TELOPT_TTYPE:
		sendopt(WILL, *s);
		break;
	     case TELOPT_NAWS:
		sendopt(WILL, *s);
		sending_size = 1;
		tcp_write_tty_size();
		break;
	     default:
		/* accept nothing else */
		sendopt(WONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;
	    
	 case GOTDONT:
#ifdef TELOPTS
	    tty_printf("[got DONT %s]\n", TELOPTSTR(*s));
#endif
	    if (*s == TELOPT_SGA) {
		linemode &= ~LM_CHAR;
		tty_special_keys();
	    }
	    sendopt(WONT, *s);
	    state = NORMAL;
	    break;
	    
	 case GOTSB:
	    if (*s == IAC) {
		state = GOTSBIAC;
	    } else {
		if (subchars < MAXSUBOPT)
		    subopt[subchars++] = *s;
	    }
	    break;
	    
	 case GOTSBIAC:
	    if (*s == IAC) {
		if (subchars < MAXSUBOPT)
		    subopt[subchars++] = IAC;
		state = GOTSB;
	    } else if (*s == SE) {
		subopt[subchars] = '\0';
		dosubopt((char*)subopt);
		state = NORMAL;
	    } else {
		/* problem! I haven't the foggiest idea of what to do here.
		 * I'll just ignore it and hope it goes away. */
		tty_printf("#telnet: got IAC <%d> instead of IAC SE!\n", (int)*s);
		state = NORMAL;
	    }
	    break;
	}
    }
    sock_state[tcp_fd] = state;
    return (char*)p - buffer;
}

/*
 * send an option negotiation
 * 'what' is one of WILL, WONT, DO, DONT
 */
static void sendopt(what, opt)
int what; unsigned opt;
{
    static unsigned char buf[3] = { IAC, 0, 0 };
    buf[1] = what; buf[2] = opt;
    
    if (write(tcp_fd, buf, 3) < 3)
	
	syserr("write option");
#ifdef TELOPTS
    tty_printf("[sent %s %s]\n", (what == WILL) ? "WILL" :
	       (what == WONT) ? "WONT" :
	       (what == DO) ? "DO" : (what == DONT) ? "DONT" : "error",
	       TELOPTSTR(opt));
#endif
}

/*
 * Send current terminal size (RFC 1073)
 */
void tcp_write_tty_size()
{
    int i;
    if (sending_size) {
	static unsigned char buf[9] =
	{ IAC, SB, TELOPT_NAWS, 0, 0, 0, 0, IAC, SE };
	
	buf[3] = cols >> 8;
	buf[4] = cols & 0xff;
	buf[5] = lines >> 8;
	buf[6] = lines & 0xff;
	
	while ((i = write(tcp_main_fd, buf, 9)) < 0 && errno == EINTR)
	    ;
	if (i != 9)
	    syserr("write term size");

#ifdef TELTOPS
	tty_printf("[term size %d %d]\n", cols, lines);
#endif
    }
}

/*
 * process suboptions.
 * so far, only terminal type is processed but future extensions are
 * window size, X display location, etc.
 */
static void dosubopt(str)
char *str;
{
    char buf[256], *term;
    
    if (str[0] == TELOPT_TTYPE) {
	if (str[1] == 1) {
	    /* 1 == SEND */
#ifdef TELOPTS
	    tty_printf("[got SB TERMINAL TYPE SEND]\n");
#endif
	    if (!(term = getenv("TERM"))) term = "unknown";
	    sprintf(buf, "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE, 0,
		    term, IAC, SE);	/* 0 == IS */
	    
	    if (write(tcp_fd, buf, strlen(term) + 6) < strlen(term) + 6)
		
		syserr("write subopt to socket");
#ifdef TELOPTS
	    tty_printf("[sent SB TERMINAL TYPE IS %s]\n", term);
#endif
	}
    }
}


static char output_buffer[OUTBUFSIZE];
static int output_len = 0;	/* number of characters in output_buffer */
static int output_socket = -1;	/* to which socket buffer should be sent*/

/*
 * put data in the output buffer for transmission to the remote host
 */
void tcp_write(data)
char *data;
{
    int len, space;
    sent += len = strlen(data);
    internal = -1;
    
    if (!(linemode & LM_NOECHO) && capfile)
        fprintf(capfile, "%s\n", data);
    
    if (tcp_fd != output_socket) { /* is there data to another socket? */
	if (output_socket >= 0)
	    tcp_flush(); /* then flush it */
	output_socket = tcp_fd;
    }
    space = OUTBUFSIZE - output_len;
    if (space <= len) {  		/* not enough room? */
	bcopy(data, output_buffer + output_len, space);
	output_len = OUTBUFSIZE;
	tcp_flush();
	tcp_write(data + space);
    } else {
        data[len] = '\n';
	bcopy(data, output_buffer + output_len, len + 1);
	output_len += len + 1;
        data[len] = '\0';
    }
    
}

/*
 * send a string to the main connection on the remote host
 */
void tcp_main_write(str)
char *str;
{
    int otcp_fd = tcp_fd;  /* backup current socket fd */
    tcp_fd = tcp_main_fd;
    tcp_write(str);
    tcp_fd = otcp_fd;
}

/*
 * send all buffered data to the remote host
 */
void tcp_flush()
{
    int n;
    char *p = output_buffer;
    
    if (output_len && output_socket == -1) {
	tty_clear_input_line(1);
	PRINT_INT ("%s#no open connections. Use '#connect main <address> <port>' to open a connection.\n%s", edattrend, edattrbeg);
	output_len = 0;
	showprompt();
	tty_draw_input_line();
	return;
    }
    
    while (output_len) {
	
	if ((n = write(output_socket, p, output_len)) < 0)
	    
	    syserr("write to socket");
	p += n;
	output_len -= n;
    }
}

/*
 * Below are multiple-connection support functions:
 */

/*
 * return connection's index or socket descriptor given id,
 * or -1 if null or invalid id is given
 */
int tcp_find (id)
char *id;
{
    int i, n;
    
    for (i=0, n=tcp_count; n; i++)
      if (CONN_LIST(i).id) 
    {
        n--;
        if (!strcmp(CONN_LIST(i).id, id))
	  return i;
    }
    return -1;
}



void show_host()
{
    if (!internal) internal = 1;
    if (*hostname)
      tty_printf("#host: %s %s\n", hostname, portnumber);
    else
      tty_puts("#host not defined.\n");
}


/*
 * show list of open connections
 */
void tcp_show()
{
    int i, n = tcp_count;
    
    PRINT_INT ("#%s connection%s opened%c\n", n ? "The following" : "No", 
	       n==1 ? " is" : "s are", n ? ':' : '.');
    
    
    for (i=0, n=tcp_count; n && i < MAXFDSCAN; i++)
      if (CONN_LIST(i).id) {
	  n--;
	  tty_printf("%s ##%s   \t%sactive \t(%s %d)\n",
		     i == tcp_main_fd ? "(default)" : "         ",
		     CONN_LIST(i).id, CONN_LIST(i).active ? "   " : "non",
		     CONN_LIST(i).host, CONN_LIST(i).port);
      }
}

/*
 * open another connection
 */
void tcp_open (id, initstr, host, port)
char *id;        /* connection id */
char *initstr;        /* optional init string */
char *host; int port;  /* address of remote host */
{
    int newtcp_fd;
    
    if (tcp_find(id)>=0) {
	PRINT_INT ("#connection `%s' already open.\n", id);
	return;
    }
    
    /* dial the number by moving the right index in small circles */
    if ((newtcp_fd = tcp_connect(host, port)) < 0) {
        PRINT_INT("#unable to connect to host. Giving up.\n");
	return;
    }
    if (echo_int && tcp_count) {
	PRINT_INT("#default connection is now `%s'\n", id);
    }
    
    FD_SET(newtcp_fd, &fdset);	       /* add socket to select() set */
    tcp_count++;
    tcp_main_fd = tcp_fd = newtcp_fd;    /* ALWAYS set as main connection */
    linemode = 0;		       /* GH: reset linemode and prompt */
    *promptstr = promptlen = 0;	       /*     when main conn changes */

    CONN_LIST(newtcp_fd).active = 1;
    CONN_LIST(newtcp_fd).id = str_dup(id);
    CONN_LIST(newtcp_fd).host = str_dup(host);
    CONN_LIST(newtcp_fd).port = port;
    
    gettimeofday(&now, NULL); /* an #at or #in may be inside initstr */
    
    if (initstr)
      parse_instruction(initstr, 0, 0);
}

/*
 * close a connection
 */
void tcp_close (id)
char *id;
{
    int i, sfd;
    
    if (id) {  /* #disconnect cmd */
	if ((sfd = tcp_find(id))<0) {
	    PRINT_INT("#no such connection: %s\n", id);
	    return;
	}
	close(sfd);
    } else
      sfd = tcp_fd;  /* connection closed by remote host */
    
    PRINT_INT("#connection on %s closed.\n", CONN_LIST(sfd).id);
    
    if (sfd == tcp_main_fd) { /* main connection closed */
	if (tcp_count == 1) { /* was last connection */
	    if (opt_exit)
		exitpowwow();
	    tty_puts("#no connections left. Type #quit to quit.\n");
	    tcp_fd = tcp_main_fd = -1;
	} else {
	    /* must find another connection and promote it to main */
	    for (i = 0; i < MAXFDSCAN; i++) {
		if (i == sfd)
		  continue;
		if (CONN_LIST(i).id) {
		    tty_printf("#default connection is now `%s'\n", CONN_LIST(i).id);
		    tcp_main_fd = i;
		    if (tcp_fd == sfd)
		      tcp_fd = -1;
		    break;
		}
	    }
	    if (sfd == tcp_main_fd) {
		tty_printf("#internal error in tcp_close()\nGiving up.\n");
		syserr(NULL);
	    }
	}
	linemode = 0;                 /* GH: reset linemode and prompt */
	*promptstr = promptlen = 0;   /*     when main conn changes */
    }
    
    if (tcp_fd == sfd)
      tcp_fd = -1; /* no further I/O allowed on sfd, as we just closed it */
    
    FD_CLR(sfd, &fdset);
    tcp_count--;
    CONN_LIST(sfd).active = 0;
    free(CONN_LIST(sfd).host);
    CONN_LIST(sfd).host = 0;
    free(CONN_LIST(sfd).id);
    CONN_LIST(sfd).id = 0;
}

/*
 * toggle output display from another connection
 */
void tcp_togglesnoop (id)
char *id;
{
    int sfd;
    
    sfd = tcp_find(id);
    if (sfd>=0) {
        CONN_LIST(sfd).active ^= 1;
	if (echo_int) {
	    PRINT_INT("#connection %s is now %sactive.\n",
		      CONN_LIST(sfd).id, CONN_LIST(sfd).active ? "" : "non");
	}
    } else {
        PRINT_INT("#no such connection: %s\n", id);
    }
}
